
# -- import packages: --------------------------------------------------------------------
from torch_adata import LightningAnnDataModule
import pandas as pd
import numpy as np
import anndata
import torch
import os
import ABCParse


# -- import local dependencies: ----------------------------------------------------------
from .. import utils
from ... import tools

from ._configure_time_key import TimeKeyConfiguration


# -- set typing: -------------------------------------------------------------------------
from typing import Optional, Dict

## --- OVERVIEW --- ##
# TimeKey
# TimeConfiguration
# configure_time
# LightningData
# DataConfiguration
## ---------------- ##

# class TimeZeroIdx(ABCParse.ABCParse):
#     def __init__(self, *args, **kwargs):
#         self.__parse__(locals())

#     @property
#     def _TIME_KEY_CONFIG(self):
#         if not hasattr(self, "_time_key_config"):
#             self._time_key_config = (
#                 sdq.core.configs._configure_time_key.TimeKeyConfiguration()
#             )
#         return self._time_key_config

#     @property
#     def _TIME_KEY(self):
#         return self._TIME_KEY_CONFIG(self._adata, time_key=self._time_key)

#     @property
#     def _TIME(self):
#         return self._adata.obs[self._TIME_KEY]

#     @property
#     def _T_MIN(self):
#         return self._TIME.min()

#     @property
#     def _HAS_TIME_KEY(self):
#         return self._TIME_KEY_CONFIG._VALID

#     def _from_time_key(self):
#         return self._adata.obs[self._TIME == self._T_MIN].index

#     @property
#     def _T0_IDX_PASSED(self):
#         return hasattr(self, "_t0_idx")

#     def __call__(self, t0_idx, *args, **kwargs):
#         self.__update__(locals())

#         if self._T0_IDX_PASSED:
#             return self._t0_idx

#         if self._HAS_TIME_KEY:
#             return self._from_time_key()
        
#         if (not self._time_cluster_key is None) and (not self._t0_cluster is None):
            
#             return self.adata.obs[
#                 self.adata.obs[self._time_cluster_key] == self._t0_cluster
#             ].index

# class TimeKey:
#     def __init__(
#         self,
#         auto_detected_time_cols=[
#             "t",
#             "Time point",
#             "time",
#             "time_pt",
#             "t_info",
#             "time_info",
#         ],
#     ):
#         self._auto_detected_time_cols = auto_detected_time_cols

#     @property
#     def _OBS_COLS(self):
#         return self._adata.obs.columns

#     @property
#     def _PASSED(self) -> bool:
#         return not (self._time_key is None)

#     @property
#     def _PASSED_VALID(self) -> bool:
#         return self._time_key in self._OBS_COLS

#     @property
#     def _DETECTED(self) -> bool:
#         return self._OBS_COLS[
#             [col in self._auto_detected_time_cols for col in self._OBS_COLS]
#         ].tolist()

#     @property
#     def _DETECTED_VALID(self) -> bool:
#         return len(self._DETECTED) == 1

#     @property
#     def _VALID(self):
#         return any([self._PASSED_VALID, self._DETECTED_VALID])

#     def __call__(self, adata: anndata.AnnData, time_key: Optional[str] = None) -> str:
        
#         self._time_key = time_key
#         self._adata = adata

#         if self._PASSED:
#             if self._PASSED_VALID:
#                 return self._time_key
#             else:
#                 raise KeyError(
#                     f"Passed `time_key`: {self._time_key} not found in adata.obs"
#                 )

#         elif self._DETECTED:
#             if self._DETECTED_VALID:
#                 return self._DETECTED[0]
#             else:
#                 print(
#                     f"More than one possible time column inferred: {found}\nPlease specify the desired time column in adata.obs."
#                 )
#         return "t"

class TimeConfiguration(ABCParse.ABCParse):
    def __init__(
        self,
        adata: anndata.AnnData,
        time_key: Optional[str] = None,
        t_min: float = 0,
        t_max: float = 1,
        dt: float = 0.1,
        t0_idx: Optional[pd.Index] = None,
        t0_cluster: Optional[str] = None,
        time_cluster_key: Optional[str] = None,
    ):
        super().__init__()
        self.__parse__(kwargs=locals(), public=["adata"])
        
        self._time_key_config = TimeKeyConfiguration()
        self._time_key = self._time_key_config(adata = self.adata, time_key = self._time_key)

    @property
    def t0_idx(self):

        """
        If t0_idx is provided, returns that.
        If t0_idx is not provided, but a cluster key and t0 cluster
        value-pair is provided, this can isolate the t0_idx.
        """

        if self.has_time_key:
            
            return self.adata.obs[
                self.adata.obs[self._time_key] == self.adata.obs[self._time_key].min()
            ].index

        if not self._t0_idx is None:
            return self._t0_idx

        if (not self._time_cluster_key is None) and (not self._t0_cluster is None):
            
            return self.adata.obs[
                self.adata.obs[self._time_cluster_key] == self._t0_cluster
            ].index

    def time_sampling(self):
        if not self.has_time_key:
            self.adata.obs["t0"] = self.adata.obs.index.isin(self.t0_idx)
#         else:
        tools.time_free_sampling(
            adata=self.adata,
            t0_idx=self.t0_idx,
            n_steps=self.n_steps,
            t0_key="t0",
            t_key="t",
        )

    @property
    def has_time_key(self):
        return self._time_key_config._VALID
        
#         return (not isinstance(self._time_key, NoneType)) and (
#             self._time_key in self.adata.obs_keys()
#         )

    @property
    def time_key(self):
        return self._time_key

#         time_key = time_key_id(adata, time_key=self._time_key)
#         if self.has_time_key:
#             return self._time_key
#         return "t"

    @property
    def t_min(self):
        if self.has_time_key:
            self._t_min = self.adata.obs[self._time_key].min()
        return self._t_min

    @property
    def t_max(self):
        if self.has_time_key:
            self._t_max = self.adata.obs[self._time_key].max()
        return self._t_max

    @property
    def t_diff(self):
        return self._t_max - self._t_min

    def _IS_CONTIGUOUS(self, n_steps):
        return n_steps % 1 == 0

    @property
    def n_steps(self):
        _n_steps = self.t_diff / self._dt + 1
        if self._IS_CONTIGUOUS(_n_steps):
            return int(_n_steps)
        raise ValueError("Time is non-contiguous")

    @property
    def dt(self):
        return self._dt

    def _t_from_time_key(self):
        return torch.Tensor(sorted(self.adata.obs[self._time_key].unique()))

    def _t_from_bounds(self):
        return torch.linspace(self._t_min, self._t_max, self.n_steps)

    def __call__(self):

        if self.has_time_key:
            return self._t_from_time_key()
        self.time_sampling()
        return self._t_from_bounds()

    @property
    def attributes(self):
        self._ATTR_KEYS = sorted(
            ["time_key", "t0_idx", "t_min", "t_max", "t_diff", "dt", "n_steps"]
        )
        return {attr: getattr(self, attr) for attr in self._ATTR_KEYS}

    def __repr__(self):

        header = "Time Attributes:\n\n  "

        _attrs = [
            "{:<9} :  {}".format(attr, val) for attr, val in self.attributes.items()
        ]
        return header + "\n  ".join(_attrs)


def configure_time(
    adata: anndata.AnnData,
    time_key: str = None,
    t_min: float = 0,
    t_max: float = 1,
    t0_idx: Optional[pd.Index] = None,
    dt: float = 0.1,
    t0_cluster: Optional[str] = None,
    time_cluster_key: Optional[str] = None,
) -> Dict:

    """Updates time for AnnData. Returns time_attributes dictionary."""

    time_config = TimeConfiguration(
        adata=adata,
        time_key=time_key,
        t_min=t_min,
        t_max=t_max,
        t0_idx=t0_idx,
        dt=dt,
        t0_cluster=t0_cluster,
        time_cluster_key=time_cluster_key,
    )
    
    t = time_config()
    
    return t, time_config


class LightningData(LightningAnnDataModule, ABCParse.ABCParse):
    def __init__(
        self,
        adata=None,
        h5ad_path=None,
        batch_size=2000,
        num_workers=os.cpu_count(),
        train_val_split=[0.8, 0.2],
        use_key="X_pca",
        obs_keys=[],
        weight_key='W',
        groupby: Optional[str] = None,
        train_key="train",
        val_key="val",
        test_key="test",
        predict_key="predict",
        silent=True,
        shuffle_time_labels = False,
        **kwargs,
    ):
        super(LightningData, self).__init__()
        
                    
        self.__parse__(locals(), public=[None])
        self._format_sinkhorn_weight_key()
        self._format_train_test_exposed_data()
        self.configure_train_val_split()
        

    @property
    def n_dim(self):
        if not hasattr(self, "_n_dim"):
            self._n_dim = self.train_dataset.X.shape[-1]
        return self._n_dim

    
    def _format_sinkhorn_weight_key(self):
        if not self._weight_key in self._adata.obs.columns:
            self._adata.obs[self._weight_key] = 1
        self._obs_keys.append(self._weight_key)
    
    def _format_train_test_exposed_data(self):
        
        if not self._train_key in self._adata.obs.columns:
            self._adata.obs[self._train_key] = True
        if not self._test_key in self._adata.obs.columns:
            self._adata.obs[self._test_key] = False
            
    def prepare_data(self):
        ...

    def setup(self, stage=None):
        ...


class DataConfiguration(ABCParse.ABCParse):
    """Should be called during config steps of scDiffEq model instantiation."""

    def __init__(self, *args, **kwargs):
        self.__parse__(locals(), public=[None])

    def _configure_anndata(self, adata):
        """
        Copy the object so as not to modify the input object, inplace.
        Stash input indices incase they are changed. Ensure that the obs
        index are str type.
        """

        # so as not to modify the input object, inplace.
        self.adata = adata.copy()

        # stash these in case they're needed later
        self._INPUT_OBS_IDX = self.adata.obs.index
        self._INPUT_VAR_IDX = self.adata.var.index
        # make sure index is str
        if not isinstance(adata.obs.index[0], str):
            self.adata = utils.idx_to_int_str(self.adata)
        self._PARAMS["adata"] = self.adata

    @property
    def _TIME_KWARGS(self) -> Dict:
        return utils.extract_func_kwargs(configure_time, self._PARAMS)

    def _configure_time(self) -> None:
        self.t, self.t_config = configure_time(**self._TIME_KWARGS)
        self._scDiffEq_kwargs["groupby"] = self.t_config.attributes["time_key"]
        self._scDiffEq_kwargs.update(self.t_config.attributes)

    @property
    def _LIT_DATA_KWARGS(self) -> Dict:
        self._PARAMS.update(self._scDiffEq_kwargs)
        return utils.extract_func_kwargs(func=LightningData, kwargs=self._PARAMS)

    def _configure_lightning_data(self):
        self.LitDataModule = LightningData(**self._LIT_DATA_KWARGS)
        self._scDiffEq_kwargs["data_dim"] = self.LitDataModule.n_dim

    def _update_scDiffEq(self, scDiffEq):
        scDiffEq._PARAMS.update(self._scDiffEq_kwargs)

        scDiffEq.t = self.t
        scDiffEq.t_config = self.t_config
        scDiffEq.LitDataModule = self.LitDataModule
        scDiffEq.adata = self.adata
        scDiffEq._data_dim = self._scDiffEq_kwargs["data_dim"]

    def __call__(self, scDiffEq, *args, **kwargs) -> None:

        """
        Data configuration occurs in four steps:
        (1) input adata is configured
        (2) Time is configured
        (3) Lightning Data is configured
        (4) scDiffEq model is updated with new attributes.
        
        Parameters
        ----------
        scDiffEq
            Derived from this are:
                adata: anndata.AnnData
                scDiffEq_kwargs: Dict
        
        Returns
        -------
        None, updates scDiffEq model.
        """
        
        adata = scDiffEq.adata
        scDiffEq_kwargs = scDiffEq._PARAMS
        
        self.__update__(locals(), ignore=["adata", "scDiffEq"], public=[None])

        self._PARAMS.update(scDiffEq_kwargs)

        self._configure_anndata(adata=adata)
        self._configure_time()
        self._configure_lightning_data()
        self._update_scDiffEq(scDiffEq)


# # -- import packages: -----------------------------------
# from torch_adata import LightningAnnDataModule
# import ABCParse
# import os


# class LightningData(LightningAnnDataModule, ABCParse.ABCParse):
#     def __init__(
#         self,
#         adata=None,
#         h5ad_path=None,
#         batch_size: int = 2000,
#         num_workers=os.cpu_count(),
#         train_val_split=[0.8, 0.2],
#         use_key: str = "X_pca",
#         obs_keys=[],
#         weight_key='W',
#         groupby="Time point",  # TODO: make optional
#         train_key="train",
#         val_key="val",
#         test_key="test",
#         predict_key="predict",
#         silent=True,
#         shuffle_time_labels = False,
#         **kwargs,
#     ):
#         super(LightningData, self).__init__()
        
                    
#         self.__parse__(locals())
#         self._format_sinkhorn_weight_key()
#         self._format_train_test_exposed_data()
#         self.configure_train_val_split()
        

#     @property
#     def n_dim(self):
#         if not hasattr(self, "_n_dim"):
#             self._n_dim = self.train_dataset.X.shape[-1]
#         return self._n_dim
    
# #     def shuffle_t(self):
        
# #         df = self._adata.obs.copy()
# #         non_t0 = df.loc[df['t'] != 0]['t']
        
# #         shuffled_t = np.zeros(len(df))
# #         shuffled_t[non_t0.index.astype(int)] = np.random.choice(non_t0.values, len(non_t0))
# #         self._adata.obs["t"] = shuffled_t
    
#     def _format_sinkhorn_weight_key(self):
#         if not self._weight_key in self._adata.obs.columns:
#             self._adata.obs[self._weight_key] = 1
#         self._obs_keys.append(self._weight_key)
    
#     def _format_train_test_exposed_data(self):
        
#         if not self._train_key in self._adata.obs.columns:
#             self._adata.obs[self._train_key] = True
#         if not self._test_key in self._adata.obs.columns:
#             self._adata.obs[self._test_key] = False
            
#     def prepare_data(self):
#         ...

#     def setup(self, stage=None):
#         ...




# -- import packages: ----------------------------------------------------------
from neural_diffeqs import LatentPotentialODE
import torch_nets
import torch


# -- import local dependencies: ------------------------------------------------
from . import mix_ins
from . import base


from typing import Optional, Union, List
from ... import __version__


# -- lightning model: ----------------------------------------------------------
class LightningODE_VAE_PriorPotential(
    mix_ins.DriftPriorMixIn,
    mix_ins.PotentialMixIn,
    mix_ins.VAEMixIn,
    mix_ins.PreTrainMixIn,
    base.BaseLightningDiffEq,
):
    def __init__(
        self,
        data_dim,
        latent_dim: int = 50,
        name: Optional[str] = None,
        train_lr=1e-5,
        pretrain_lr=1e-3,
        pretrain_epochs=100,
        pretrain_optimizer=torch.optim.Adam,
        train_optimizer=torch.optim.RMSprop,
        pretrain_scheduler=torch.optim.lr_scheduler.StepLR,
        train_scheduler=torch.optim.lr_scheduler.StepLR,
        pretrain_step_size=200,
        train_step_size=10,
        dt=0.1,
        adjoint=False,
        backend = "auto",
        
        # -- encoder parameters: -------
        encoder_n_hidden: int = 4,
        encoder_power: float = 2,
        encoder_activation: Union[str, List[str]] = 'LeakyReLU',
        encoder_dropout: Union[float, List[float]] = 0.2,
        encoder_bias: bool = True,
        encoder_output_bias: bool = True,

        # -- decoder parameters: -------
        decoder_n_hidden: int = 4,
        decoder_power: float = 2,
        decoder_activation: Union[str, List[str]] = 'LeakyReLU',
        decoder_dropout: Union[float, List[float]] = 0.2,
        decoder_bias: bool = True,
        decoder_output_bias: bool = True,
        
        version = __version__,
        
        *args,
        **kwargs,
    ):
        super().__init__()
        
        name = self._configure_name(name)

        self.save_hyperparameters()
        
        # -- torch modules: ----------------------------------------------------
        self._configure_torch_modules(func = LatentPotentialODE, kwargs=locals())
        self._configure_lightning_model(kwargs = locals())

    def log_computed_loss(self, sinkhorn_loss, t, kl_div_loss, stage):
        
        sinkhorn_loss = self.log_sinkhorn_divergence(sinkhorn_loss).sum()
        self.log(f"kl_div_{stage}", kl_div_loss.sum())

        return sinkhorn_loss + kl_div_loss.sum()

    def step(self, batch, batch_idx, stage=None):

        batch = self.process_batch(batch, batch_idx)
        X_hat, kl_div_loss = self.forward(batch.X0, batch.t)
        sinkhorn_loss = self.compute_sinkhorn_divergence(
            batch.X, X_hat, batch.W, batch.W_hat
        )
        return self.log_computed_loss(
            sinkhorn_loss, t=batch.t, kl_div_loss=kl_div_loss, stage=stage
        )

    def pretrain_step(self, batch, batch_idx, stage=None):

        batch = self.process_batch(batch, batch_idx)
        X0_hat = self.Decoder(self.Encoder(batch.X0))
        recon_loss = self.reconstruction_loss(X0_hat, batch.X0).sum()
        self.log("pretrain_rl_mse", recon_loss.item())
        return recon_loss

    def __repr__(self):
        return "LightningODE-VAE-PriorPotential"



# -- import packages: ----------------------------------------------------------
import anndata
import ABCParse
import adata_query
import pandas as pd


# -- import local dependencies: ------------------------------------------------
from ..core import utils


# -- set typing: ---------------------------------------------------------------
from typing import Union, List, Dict


# -- controller class: ---------------------------------------------------------
class GroupedExpression(ABCParse.ABCParse):
    def __init__(
        self,
        adata: anndata.AnnData,
        gene_id_key: str = "gene_ids",
        use_key: str = "X_scaled",
        *args,
        **kwargs,
    ):
        """

        Parameters
        ----------
        adata: anndata.AnnData

        gene_id_key: str
            "gene_ids"

        use_key: str
            "X_scaled"

        Returns
        -------
        None

        Notes
        -----
        1.  If simulated AnnData is passed, you must have gene_id_key in adata.uns,
            i.e.,: in adata.uns[gene_id_key].
        """

        self.__parse__(locals(), ignore=["adata"], public=[None])
        self._process_adata(adata)

    @property
    def _SIMULATED_DETECTED(self):
        return self._gene_id_key in self.adata.uns

    def _process_simulated_anndata(self):
        """
        Updates self.adata and self._use_key
        """
        
        X_gene = adata_query.fetch(self.adata, key=self._use_key, torch=False)
        
        if isinstance(X_gene, pd.DataFrame):
            X_gene = X_gene.values

        var_names = self.adata.uns[self._gene_id_key]
        self.adata = anndata.AnnData(
            X=X_gene,
            dtype=X_gene.dtype,
            var=var_names.to_frame(),
            obs=self.adata.obs,
        )
        self._use_key = "X"

    def _process_adata(self, adata):

        self.adata = adata.copy()

        if self._SIMULATED_DETECTED:
            self._process_simulated_anndata()

        self.adata.var_names = self.adata.var[self._gene_id_key]

    def _single_gene_expression(self, df):
        return adata_query.fetch(
            adata = self.adata[df.index, self._gene_id],
            key=self._use_key,
            torch=False,
        )

    @property
    def _GROUPED(self):
        return self.adata.obs.groupby(self._groupby)

    def _process_output(self):
        return {
            self._gene_id: self._GROUPED.apply(self._single_gene_expression).to_dict()
        }

    def __call__(self, gene_id: str, groupby: str) -> Dict:
        """

        Parameters
        ----------
        gene_id: str

        groupby: str

        Returns
        -------
        grouped_expression: Dict

        Notes
        -----
        """

        self.__update__(locals(), private=["gene_id", "groupby"])

        return self._process_output()


# -- API-facing function: ------------------------------------------------------
def grouped_expression(
    adata: anndata.AnnData,
    gene_id: Union[List[str], str],
    groupby: str,
    gene_id_key: str = "gene_ids",
    use_key: str = "X_scaled",
    *args,
    **kwargs,
) -> Dict:

    """
    Parameters
    ----------
    adata: anndata.AnnData
        AnnData object.

    gene_id: Union[List[str], str]
        Gene or list of genes for which to obtain grouped expression values.

    groupby: str
        obs_key by which to group cells.

    gene_id_key: str
        var_key to set var_names for fetching gene_ids.

    use_key: str
        Key to grab a specific matrix from AnnData.

    Returns
    -------
    grouped_expression: Dict

        Organized as:
        {
            gene_1: {
                group1: np.ndarray(...),
                group2: np.ndarray(...),
                ...
            },
            gene_2: {
                group1: np.ndarray(...),
                group2: np.ndarray(...),
                ...
            },
        }

    Notes
    -----
    1.  If simulated AnnData is passed, you must have gene_id_key in adata.uns,
        i.e.,: in adata.uns[gene_id_key].
    """

    if isinstance(gene_id, str):
        gene_id = [gene_id]

    grouped_expr = GroupedExpression(adata, gene_id_key=gene_id_key, use_key=use_key)
    return {_id: grouped_expr(_id, groupby)[_id] for _id in gene_id}




__module_name__ = "__init__.py"
__doc__ = """I/O __init__ module. Sub-package of the main scdiffeq API."""
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["mvinyard.ai@gmail.com"])


# import functions accessed as sdq.io.<func>: --------------------------------------------
from ._read_h5ad import read_h5ad
from ._pickle_io import read_pickle, write_pickle
from ._model_loader import load_diffeq, load_model



import numpy as np


from ._annotate_cells import annotate_cells


class TimeFreeSampling:
    def __init__(self, adata, n_steps, t0_key="t0"):
        self.__parse__(locals())
        
    def __parse__(self, kwargs, ignore=['self']):
        
        for key, val in kwargs.items():
            if not key in ignore:
                setattr(self, key, val)        
        
    def annotate_t0(self, t0_idx):
        annotate_cells(self.adata, idx=t0_idx, key=self.t0_key)
        
    @property
    def non_t0_idx(self):
        df = self.adata.obs.copy()
        return df.loc[~df[self.t0_key]].index
    
    @property
    def stepwise_samples(self):
        return np.random.choice(range(1, self.n_steps), self.non_t0_idx.shape)
    
    @property
    def time_free_sampled_groups(self):
        idx = np.zeros(len(self.adata))
        # organize remaining non-t0 cells into n_steps # of groups
        idx[self.non_t0_idx.astype(int)] = self.stepwise_samples
        return idx
    
def time_free_sampling(adata, t0_idx, n_steps, t0_key="t0", t_key="t"):
    """
    Notes:
    ------
    Could potentially improve this through a probabalistic sampling
    processe, weighting certain cells to be more likely assigned to
    'later' groups.
    """
    tfs = TimeFreeSampling(adata, n_steps=n_steps, t0_key=t0_key)
    tfs.annotate_t0(t0_idx)
    adata.obs[t_key] = tfs.time_free_sampled_groups
    adata.obs[t_key] = (adata.obs[t_key] / adata.obs[t_key].max())





from neural_diffeqs import NeuralSDE


def default_NeuralSDE(
    state_size,
    mu_hidden=[1600, 1600],
    sigma_hidden=[800, 800],
    mu_activation="LeakyReLU",
    sigma_activation="LeakyReLU",
    mu_dropout=0.2,
    sigma_dropout=0.2,
    mu_bias=True,
    sigma_bias=True,
    mu_output_bias=True,
    sigma_output_bias=True,
    mu_potential=False,
    sigma_potential=False,
    noise_type="general",
    sde_type="ito",
    brownian_size=1,

):
    return NeuralSDE(
        state_size=state_size,
        mu_hidden=mu_hidden,
        sigma_hidden=sigma_hidden,
        mu_activation=mu_activation,
        sigma_activation=sigma_activation,
        mu_dropout=mu_dropout,
        sigma_dropout=sigma_dropout,
        mu_bias=mu_bias,
        sigma_bias=sigma_bias,
        mu_output_bias=mu_output_bias,
        sigma_output_bias=sigma_output_bias,
        mu_potential=mu_potential,
        sigma_potential=sigma_potential,
        noise_type=noise_type,
        sde_type=sde_type,
        brownian_size=brownian_size,
    )



__module_name__ = "__init__.py"
__doc__ = """TODO"""
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu"])


# -- import: -------------------------------------------------------------------
from ._function_fetch import FunctionFetch, fetch_optimizer, fetch_lr_scheduler
from ._function_kwargs import extract_func_kwargs, function_kwargs
from ._logging_learnable_hparams import LoggingLearnableHParams
from ._sum_normalize import sum_normalize

from ._scdiffeq_logger import scDiffEqLogger
from ._logger_bridge import LoggerBridge


from ._normalize_time_to_range import normalize_time

from ._not_none_type import not_none


from ._anndata_inspector import AnnDataInspector
from ._info_message import InfoMessage
from ._fast_graph import FastGraph


from ._idx_to_int_str import idx_to_int_str

from ._logs import Logs, PretrainLogs, TrainLogs

from ._filter_df import filter_df

from ._display_tracked_loss import display_tracked_loss

from ._knn_graph_query import kNNGraphQuery
from ._flexible_component_loader import FlexibleComponentLoader


# review for deletion:
# from ._default_neural_sde import default_NeuralSDE
# from ._fetch_format import fetch_format



import ABCParse
import anndata
import numpy as np
import pandas as pd

from typing import Optional, Union, Dict, List, Tuple

class TemporalExpression(ABCParse.ABCParse):
    def __init__(
        self,
        time_key: str = "t",
        groupby: Optional[Union[None, str]] = None,
        gene_id_key: str = "gene_ids",
        use_key: str = "X_gene_inv",
        *args,
        **kwargs,
    ):
        self.__parse__(locals(), public=[None])

    @property
    def _GROUPBY(self):
        """
        if no additional groupby key is passed, we just group by time.
        """
        if self._groupby is None:
            return self._time_key

        groupby = [self._time_key]
        groupby.append(self._groupby)
        return groupby

    @property
    def _GROUPED(self):
        if not hasattr(self, "_grouped"):
            self._grouped = self._adata_sim.obs.groupby(self._GROUPBY)
        return self._grouped

    @property
    def _GENE(self):
        if not hasattr(self, "_gene") or (self._gene is None):
            self._gene = self._adata_sim.uns[self._gene_id_key].tolist()
        return self._gene

    @property
    def _MULTIGROUPED(self):
        return isinstance(self._GROUPBY, List)

    def _forward(self, df, operation=np.mean):
        gene = self._adata_sim[df.index].obsm[self._use_key][self._GENE]
        return operation(gene)

    def _add_to_adata(self, mean, std):
        """"""
        self._adata_sim.uns["gex_mean"] = mean
        self._INFO("Mean temporal expression added to: `adata_sim.uns['gex_mean']`")
        self._adata_sim.uns["gex_std"] = std
        self._INFO(
            "Standard deviation of temporal expression added to: `adata_sim.uns['gex_std']`"
        )

    def __call__(
        self,
        adata_sim: anndata.AnnData,
        gene: Optional[Union[List[str], str]] = None,
        return_gex: bool = False,
        *args,
        **kwargs,
    ) -> Union[None, Tuple[pd.DataFrame]]:

        """
        Parameters
        ----------
        adata_sim: anndata.AnnData

        gene: Optional[Union[List[str], str]], default = None

        return_gex: bool, default = False

        Returns
        -------
        mean, std: Union[None, Tuple[pd.DataFrame]]
            If `return_gex` is True, returns values.
        """

        self.__update__(locals(), public=[None])

        mean = self._GROUPED.apply(self._forward, operation=np.mean)
        std = self._GROUPED.apply(self._forward, operation=np.std)

        if self._MULTIGROUPED:
            mean, std = mean.unstack(), std.unstack()

        self._add_to_adata(mean, std)

        if return_gex:
            return mean, std


def temporal_expression(
    adata_sim: anndata.AnnData,
    groupby: Optional[Union[None, str]] = "final_state",
    gene: Optional[Union[List[str], str]] = None,
    time_key: str = "t",
    gene_id_key: str = "gene_ids",
    use_key: str = "X_gene_inv",
    return_gex: bool = False,
) -> Union[None, Tuple[pd.DataFrame]]:
    """
    Parameters
    ----------
    adata_sim: anndata.AnnData

    groupby: Optional[Union[None, str]], default = "final_state"

    gene: Optional[Union[List[str], str]], default = None

    time_key: str, default = "t"

    gene_id_key: str, default = "gene_ids"

    use_key: str, default = "X_gene_inv"

    return_gex: bool, default = False

    Returns
    -------
    mean, std: Union[None, Tuple[pd.DataFrame]]
        If `return_gex` is True, returns values.
    """

    t_expr = TemporalExpression(
        time_key=time_key, groupby=groupby, gene_id_key=gene_id_key, use_key=use_key
    )
    return t_expr(adata_sim=adata_sim, gene=gene, return_gex=return_gex)




import os
from licorice_font import font_format
from ._versions import configure_version
from ._reconstruct_function import reconstruct_function
from ._hyperparams import HyperParams

note = font_format("NOTE", ["BLUE"])
error = font_format("FileNotFoundError", ["RED"])

def func_from_version(
    version=None,
    model_type="NeuralSDE",
    base_dir=".",
    base_path="scDiffEq_model/lightning_logs/version_{}/hparams.yaml",
):

    version, yaml_base_path, versions_available = configure_version(
        version=version, base_dir=base_dir, base_path=base_path
    ).values()
    
    yaml_path = yaml_base_path.format(version)
    if os.path.exists(yaml_path):
        try:
            hparams = HyperParams(yaml_path)
            print(" - [{}] | Reconstructing function from version: {}".format(note, version))
            return reconstruct_function(hparams)
        except:
            raise ValueError("hparams.yaml was found but there was a problem reconstruction the function.")

    fmt_available = [yaml_base_path.format(v) for v in versions_available]
    print(fmt_available)
    
    raise FileNotFoundError(
        "\n\n - [{}] | hparams.yaml path not found: {}\n\nThe following are available:\n\t{}".format(
            error, yaml_path, *fmt_available
        )
    )




import ABCParse
import umap


class UMAP(ABCParse.ABCParse):
    def __init__(
        self,
        use_key="X_pca",
        n_neighbors=15,
        metric="euclidean",
        spread=2,
        **kwargs,
    ):
        
        super(UMAP, self).__init__()

        self.__parse__(locals())
        self.__config__()
        
    @property
    def _UMAP_KWARGS(self):
        return ABCParse.function_kwargs(
            func = umap.UMAP, kwargs = self._KWARGS,
        )
        
    def __config__(self):

        self.model = umap.UMAP(**self._UMAP_KWARGS)

    def forward(self, adata):

        X_use = adata.obsm[self.use_key]
        n_cells = X_use.shape[0]
        print("Fitting UMAP model to {} cells".format(n_cells))
        return self.model.fit_transform(X_use)

    def __call__(self, adata, subset_key=None):

        self.__parse__(locals(), ignore=["adata"])
        return self.forward(adata)



# -- import packages: ----------------------------------------------------------
import scipy
import anndata
import tqdm
import pandas as pd
import ABCParse


# -- import local dependencies: ------------------------------------------------
from ..core import utils


# -- set typing: ---------------------------------------------------------------
from types import FunctionType


# -- Controller Class: ---------------------------------------------------------
class FeatureCorrelation(ABCParse.ABCParse):
    """
    Calculate the per-cell correlation of a gene feature with a column
    in `adata.obs`.
    """

    def __init__(
        self,
        adata: anndata.AnnData,
        obs_key: str,
        use_key: str = "X_scaled",
        corr_func: FunctionType = scipy.stats.pearsonr,
    ):

        self.__parse__(locals(), public=["adata"])
        self._CorrelationDict = {}

    @property
    def VARS(self):
        return self.adata.var_names

    @property
    def OBS_PROPERTY(self):
        if not hasattr(self, "_obs_property"):
            self._obs_property = self.adata.obs[self._obs_key].values
        return self._obs_property

    @property
    def _PROGRESS_BAR(self):
        return tqdm.notebook.tqdm(
            self.VARS,
            desc = f"Computing gene correlation to {self._obs_key}"
        )

    def _gene(self, gene):
        if self._use_key == "X":
            return self.adata[:, gene].X.flatten()
        else:
            return self.adata[:, gene].layers[self._use_key].flatten()

    def forward(self, gene):
        corr, pval = self._corr_func(self._gene(gene), self.OBS_PROPERTY)
        return {"corr": corr, "pval": pval}

    def _to_frame(self):
        if not hasattr(self, "_corr_df"):
            self._corr_df = pd.DataFrame(self._CorrelationDict).T
        return self._corr_df

    def _annotate_adata(self):
        self.adata.var[f"{self._obs_key}_correlation"] = self._corr_df["corr"].values
        self.adata.var[f"{self._obs_key}_correlation_pval"] = self._corr_df[
            "pval"
        ].values

    def __call__(self, add_to_adata: bool = True):

        for var in self._PROGRESS_BAR:
            self._CorrelationDict[var] = self.forward(var)

        self._to_frame()
        if add_to_adata:
            self._annotate_adata()
        else:
            return self._corr_df


def drift_correlated_features(
    adata: anndata.AnnData,
    drift_key: str = "drift",
    use_key: str = "X_scaled",
    correlation_func: FunctionType = scipy.stats.pearsonr,
    add_to_adata: bool = True,
):
    """
    Compute correlation between per-cell gene values and drift values.
    
    Parameters:
    -----------
    adata
        type: anndata.AnnData
    
    drift_key
        type: str
        default: "drift"
    
    use_key
        type: str
        default: "X_scaled"
        
    correlation_func
        type: FunctionType
        default: scipy.stats.pearsonr
        
        
    add_to_adata
        type: bool
        default: True
        
    
    Returns:
    --------
    None, optionally adata. Updates adata.
    
    Notes:
    ------
    
    """
    
    feature_correlation = FeatureCorrelation(
        adata=adata,
        obs_key=drift_key,
        use_key=use_key,
        corr_func=correlation_func,
    )

    return feature_correlation(add_to_adata=add_to_adata)

def diffusion_correlated_features(
    adata: anndata.AnnData,
    diffusion_key: str = "diffusion",
    use_key: str = "X_scaled",
    correlation_func: FunctionType = scipy.stats.pearsonr,
    add_to_adata=True,
):
    """
    Compute correlation between per-cell gene values and diffusion values.
    
    Parameters:
    -----------
    adata
        type: anndata.AnnData
    
    diffusion_key
        type: str
        default: "diffusion"
    
    use_key
        type: str
        default: "X_scaled"
        
    correlation_func
        type: FunctionType
        default: scipy.stats.pearsonr
        
        
    add_to_adata
        type: bool
        default: True
        
    
    Returns:
    --------
    None, optionally adata. Updates adata.
    
    Notes:
    ------
    
    """
    
    feature_correlation = FeatureCorrelation(
        adata=adata,
        obs_key=diffusion_key,
        use_key=use_key,
        corr_func=correlation_func,
    )

    return feature_correlation(add_to_adata=add_to_adata)

def potential_correlated_features(
    adata: anndata.AnnData,
    potential_key: str = "potential",
    use_key: str = "X_scaled",
    correlation_func: FunctionType = scipy.stats.pearsonr,
    add_to_adata=True,
):
    """
    Compute correlation between per-cell gene values and potential values.
    
    Parameters:
    -----------
    adata
        type: anndata.AnnData
    
    potential_key
        type: str
        default: "psi"
    
    use_key
        type: str
        default: "X_scaled"
        
    correlation_func
        type: FunctionType
        default: scipy.stats.pearsonr
        
        
    add_to_adata
        type: bool
        default: True
        
    
    Returns:
    --------
    None, optionally adata. Updates adata.
    
    Notes:
    ------
    
    """
    
    feature_correlation = FeatureCorrelation(
        adata=adata,
        obs_key=potential_key,
        use_key=use_key,
        corr_func=correlation_func,
    )

    return feature_correlation(add_to_adata=add_to_adata)




# from ._sde_mix_in import SDEMixIn

from ._base_forward_mix_in import BaseForwardMixIn
from ._potential_mix_in import PotentialMixIn
from ._pre_train_mix_in import PreTrainMixIn
from ._vae_mix_in import VAEMixIn
from ._drift_prior_mix_in import DriftPriorMixIn, DriftPriorVAEMixIn


# -- fate bias mix ins: ----------------------------------------------------------------
from ._fate_bias_mix_in import FateBiasMixIn
from ._fate_bias_vae_mix_in import FateBiasVAEMixIn
from ._fate_bias_drift_prior_mix_in import FateBiasDriftPriorMixIn
from ._fate_bias_drift_prior_vae_mix_in import FateBiasDriftPriorVAEMixIn




import torch


class PotentialMixIn(object):
    def __init__(self):
        super().__init__()

    @torch.enable_grad()
    @torch.inference_mode(False)
    def validation_step(self, batch, batch_idx, *args, **kwargs):
        if not self.PRETRAIN:
            return self.step(batch, batch_idx, stage="validation")



import os
import ABCParse
from lightning import Callback
from lightning.pytorch.callbacks import ModelCheckpoint, StochasticWeightAveraging

"""where built-in callbacks are configured"""

from .. import utils, callbacks


class InterTrainerEpochCounter(Callback):
    def __init__(self):
        ...
#         self.COMPLETED_EPOCHS = 0

    def on_train_epoch_end(self, trainer, pl_module, *args, **kwargs):

        pl_module.COMPLETED_EPOCHS += 1
        ce = pl_module.COMPLETED_EPOCHS

class LightningCallbacksConfiguration(ABCParse.ABCParse):
    def __init__(self):
        super().__init__()
        
        self.cbs = []

    @property
    def BuiltInCallbacks(self):

        return [
        ModelCheckpoint(
            every_n_epochs=self._every_n_epochs,
            save_on_train_epoch_end=True,
            save_top_k=self._save_top_k,
            save_last=self._save_last,
            monitor=self._monitor,
        ),
            InterTrainerEpochCounter(),
            callbacks.ModelLogging(),
#             callbacks.VisualizeTrackedLoss(
#                 **utils.extract_func_kwargs(func = callbacks.VisualizeTrackedLoss, kwargs = self._PARAMS),
#             ),
            # StochasticWeightAveraging(swa_lrs=self.swa_lrs),
            # considering rm SWA pending better understanding
        ]

    @property
    def Callbacks(self):
        return self.cbs + self.BuiltInCallbacks
        
    @property
    def GradientRetainedCallbacks(self):
        return [callbacks.GradientPotentialTest()] + self.cbs

    def __call__(
        self,
        version,
        viz_frequency = 1,
        model_name="scDiffEq_model",
        working_dir=os.getcwd(),
        train_version=0,
        pretrain_version=0,
        callbacks=[],
        ckpt_frequency=10,
        keep_ckpts=-1,
        swa_lrs=1e-5,
        monitor=None,
        retain_test_gradients=False,
        save_last=True,
    ):
        
        self.__parse__(locals(), ignore=["callbacks"])
        
        self._every_n_epochs = ckpt_frequency
        self._save_top_k = keep_ckpts
        
        [self.cbs.append(cb) for cb in callbacks]
        
        if retain_test_gradients:
            return self.GradientRetainedCallbacks
        return self.Callbacks




# -- import packages: ----------------------------------------------------------
from neural_diffeqs import NeuralSDE
import torch


# -- import local dependencies: ------------------------------------------------
from . import base, mix_ins


from typing import Optional, Union, List

from ... import __version__


# -- lightning model: ----------------------------------------------------------
class LightningSDE_VAE_FateBiasAware(
    mix_ins.VAEMixIn,
    mix_ins.FateBiasVAEMixIn,
    mix_ins.PotentialMixIn,
    mix_ins.PreTrainMixIn,
    base.BaseLightningDiffEq,
):
    def __init__(
        self,
        data_dim,
        latent_dim: int = 50,
        name: Optional[str] = None,
        train_lr=1e-5,
        pretrain_lr=1e-3,
        pretrain_epochs=100,
        pretrain_optimizer=torch.optim.Adam,
        train_optimizer=torch.optim.RMSprop,
        pretrain_scheduler=torch.optim.lr_scheduler.StepLR,
        train_scheduler=torch.optim.lr_scheduler.StepLR,
        pretrain_step_size=200,
        train_step_size=10,
        dt=0.1,
        adjoint=False,
        backend = "auto",
        
        # -- sde params: -------------------------------------------------------
        mu_hidden: Union[List[int], int] = [400, 400, 400],
        sigma_hidden: Union[List[int], int] = [400, 400, 400],
        mu_activation: Union[str, List[str]] = 'LeakyReLU',
        sigma_activation: Union[str, List[str]] = 'LeakyReLU',
        mu_dropout: Union[float, List[float]] = 0.1,
        sigma_dropout: Union[float, List[float]] = 0.1,
        mu_bias: bool = True,
        sigma_bias: List[bool] = True,
        mu_output_bias: bool = True,
        sigma_output_bias: bool = True,
        mu_n_augment: int = 0,
        sigma_n_augment: int = 0,
        sde_type='ito',
        noise_type='general',
        brownian_dim=1,
        coef_drift: float = 1.0,
        coef_diffusion: float = 1.0,
        
        # -- encoder parameters: -----------------------------------------------
        encoder_n_hidden: int = 4,
        encoder_power: float = 2,
        encoder_activation: Union[str, List[str]] = 'LeakyReLU',
        encoder_dropout: Union[float, List[float]] = 0.2,
        encoder_bias: bool = True,
        encoder_output_bias: bool = True,
        
        # -- decoder parameters: -----------------------------------------------
        decoder_n_hidden: int = 4,
        decoder_power: float = 2,
        decoder_activation: Union[str, List[str]] = 'LeakyReLU',
        decoder_dropout: Union[float, List[float]] = 0.2,
        decoder_bias: bool = True,
        decoder_output_bias: bool = True,

        # -- fate bias parameters: ---------------------------------------------
        t0_idx = None,
        kNN_Graph=None,
        fate_bias_csv_path=None,
        fate_bias_multiplier = 1,
        
        PCA = None,
        
        version = __version__,
        
        *args,
        **kwargs,
    ):
        super().__init__()

        name = self._configure_name(name)
        
        self.save_hyperparameters(ignore=['kNN_Graph'])
                        
        # -- torch modules: ----------------------------------------------------
        self._configure_torch_modules(func=NeuralSDE, kwargs=locals())
        self._configure_lightning_model(kwargs = locals())
        
        self._configure_fate(
            graph=kNN_Graph,
            csv_path = fate_bias_csv_path,
            t0_idx = t0_idx,
            fate_bias_multiplier = fate_bias_multiplier,
            PCA = PCA,
        )

    def __repr__(self):
        return "LightningSDE-VAE-FateBiasAware"



from datetime import datetime
from pathlib import Path
import os, glob
import ABCParse

from ._info_message import InfoMessage

class scDiffEqLogger(ABCParse.ABCParse):
    """
    While Lightning uses automatic logging, we need something one step removed from this to take full
    advantage of their setup within the constraints of our model.
    """

    def __init__(
        self,
        model_name="scDiffEq_model",
        ckpt_path = None,
        working_dir=os.getcwd(),
    ):

        self.__parse__(locals(), public=[None])
        self._INFO = InfoMessage()
        self.creation_count = 0

    @property
    def _MODEL_NAME(self):
        return self._model_name
    
    @property
    def _WORKING_DIR(self):
        return self._working_dir

    @property
    def _EXISTING_VERSIONS(self):
        return glob.glob(self.PARENT_MODEL_OUTDIR + "/version*")
    
    @property
    def PARENT_MODEL_OUTDIR(self):
        return os.path.join(self._WORKING_DIR, self._MODEL_NAME)
    
    @property
    def LOG_PATH(self):
        return os.path.join(self.PARENT_MODEL_OUTDIR, "scDiffEq.log")

    @property
    def VERSIONED_MODEL_OUTDIR(self):
        if not self._ckpt_path is None:
            if self._PASSED_CKPT_MATCHES_MODEL_PARENT_DIR:
                return self.VERSION_FROM_CKPT
        
        return os.path.join(
            self.PARENT_MODEL_OUTDIR,
            "version_{}".format(len(self._EXISTING_VERSIONS)),
        )
    
    def _configure_parent_model_outdir(self):
        if not os.path.exists(self.PARENT_MODEL_OUTDIR):
            os.mkdir(self.PARENT_MODEL_OUTDIR)
            line = "".join(
                [
                    "\n ------------- scDiffEq -------------\n\n\n",
                    f" ---- {datetime.now()} ----\n\n",
                ]
            )

            f = open(self.LOG_PATH, mode="w")
            f.write(line)
            f.close()
            
    def _configure_versioned_model_outdir(self):
        
        if not os.path.exists(self.VERSIONED_MODEL_OUTDIR):
            if not self.creation_count:
                os.mkdir(self.VERSIONED_MODEL_OUTDIR)
            f = open(self.LOG_PATH, mode="a")
            v_path = self.VERSIONED_MODEL_OUTDIR
            v = os.path.basename(v_path)
            line = f"\t{v}\t{datetime.now()}\t{v_path}\n"
            f.write(line)
            f.close()
            self.creation_count += 1

        elif self._PASSED_CKPT_MATCHES_MODEL_PARENT_DIR:
            if len(self.CKPT_PATH) < 65:
                self._INFO(f"Loading from checkpoint: {self.CKPT_PATH}")
            else:
                self._INFO(f"Loading from checkpoint:\n\t{self.CKPT_PATH}")
        else:
            self._INFO(f"Directory: {self.VERSIONED_MODEL_OUTDIR} already exists!")

    # -- checkpoint loading: ------------------------------------------------
    @property
    def CKPT_PATH(self):
        return Path(self._ckpt_path).absolute().as_posix()
    
    @property
    def VERSION_FROM_CKPT(self):
        return os.path.dirname(self.CKPT_PATH.split("_logs/version_")[0])
    
    @property
    def _PASSED_CKPT_MATCHES_MODEL_PARENT_DIR(self)->bool:
        compare = [self.CKPT_PATH, self.PARENT_MODEL_OUTDIR]
        common = os.path.commonprefix(compare)
        
        return (common == self.PARENT_MODEL_OUTDIR)
    
    def __call__(self):
        self._configure_parent_model_outdir()
        self._configure_versioned_model_outdir()




"""
Three main classes that are going to be used here:

1. [ called in callback ] ModelTracker
2. [ called in callback ] LossTrackingVisualization 
3. [ lightning.Callback ] VisualizeTrackedLoss
"""
import lightning
import glob
import os

from .. import utils

from typing import List
import scdiffeq_plots as sdq_pl
import matplotlib.pyplot as plt
import ABCParse


class ModelTracker(ABCParse.ABCParse):
    def __init__(
        self,
        version,
        model_name="scDiffEq_model",
        working_dir=os.getcwd(),
        train_version=0,
        pretrain_version=0,
    ):
        super().__init__()
        self.__parse__(locals(), public=[None])

    @property
    def _PATH(self):
        return os.path.join(
            self._working_dir, self._model_name, f"version_{self._version}"
        )

    def _HAS_STAGE_LOGS(self, stage_key):
        for fpath in glob.glob(os.path.join(self._PATH, "*/version_*/metrics.csv")):
            if stage_key in fpath:
                return True
            

    @property
    def _HAS_PRETRAIN(self):
        return self._HAS_STAGE_LOGS("/pretrain_logs")

    @property
    def _HAS_TRAIN(self):
        return self._HAS_STAGE_LOGS("/train_logs")

    @property
    def _PRETRAIN(self):
        if self._HAS_PRETRAIN:
            return utils.PretrainLogs(
                self._PATH, stage="pretrain", version=self._pretrain_version
            )

    @property
    def _TRAIN(self):
        if self._HAS_TRAIN:
            return utils.TrainLogs(
                self._PATH, stage="train", version=self._train_version
            )

    def _subset_df_for_plotting(
        self,
        df,
        re_patterns=(
            ["pretrain_rl_mse"],
            ["sinkhorn", "fate_weighted"],
            ["sinkhorn_fate_weighted"],
            ["total"],
            ["fate_accuracy"],
            ["kl_div"],
        ),
        include_regex=[[True], [True, False], [True], [True], [True], [True]],
    ):
        """
        Returns dictionary of dfs subset for values included on individual plots.
        Filters PlotFrames Dict to return only measured metrics
        """

        PlotFrames = {}
        for regex, include in zip(re_patterns, include_regex):
            filt_df = utils.filter_df(df, regex=regex, include=include)
            measured = filt_df.shape[1] > 0
            if measured:
                PlotFrames[regex[0]] = filt_df

        return PlotFrames

    @property
    def pretrain_df(self):
        if self._HAS_PRETRAIN:
            return self._PRETRAIN()

    @property
    def train_df(self):
        if self._HAS_TRAIN:
            return self._TRAIN()

    @property
    def pretrain_plot_inputs(self):
        if self._HAS_PRETRAIN:
            return self._subset_df_for_plotting(self.pretrain_df)

    @property
    def train_plot_inputs(self):
        if self._HAS_TRAIN:
            return self._subset_df_for_plotting(self.train_df)


class LossTrackingVisualization(ABCParse.ABCParse):
    def __init__(
        self,
        tracker,
        ncols: int = 3,
        scale: float = 1,
        width: float = 1,
        height: float = 1,
        hspace: float = 0.4,
        wspace: float = 0.4,
        width_ratios: List[float] = None,
        height_ratios: List[float] = None,
        linearize=True,
        rm_ticks=False,
        color=[None],
        move=[0],
        xy_spines: bool = True,
        delete_spines=[[]],
        color_spines=[[]],
        move_spines=[[]],
        xlabel="Epoch",
        title_fontsize=10,
        label_fontsize=8,
        tick_param_size=6,
        colors=["dodgerblue", "darkorange"],
        grid=True,
        grid_kwargs={"zorder": 0, "alpha": 0.5},
        legend_kwargs={"edgecolor": "None", "facecolor": "None", "fontsize": 8},
    ):
        super().__init__()

        self.__parse__(locals(), public=[None])
        self._configure_plot_inputs()

    def _configure_plot_inputs(self):
        self._PLOT_INPUTS = {}
        if self._HAS_PRETRAIN:
            self._PLOT_INPUTS["pretrain"] = self._tracker.pretrain_plot_inputs
        self._PLOT_INPUTS["train"] = self._tracker.train_plot_inputs

    @property
    def _PLOT_TITLES(self):
        return {
            "pretrain_rl_mse": "Pretrain Reconstruction Loss (MSE)",
            "sinkhorn": "Sinkhorn Divergence",
            "sinkhorn_fate_weighted": "Fate-Weighted Sinkhorn Divergence",
            "fate_accuracy": "Fate Accuracy",
            "kl_div": "KL Divergence",
            "total": "Total Loss",
        }

    @property
    def _PLOT_LABELS(self):
        return {
            "pretrain_rl_mse": "MSE",
            "sinkhorn": "Wasserstein Distance",
            "sinkhorn_fate_weighted": "Wasserstein Distance",
            "fate_accuracy": "Accuracy Score",
            "kl_div": "KL Divergence",
            "total": "Mixed Units",
        }

    @property
    def _PLOT_KWARGS(self):
        return utils.extract_func_kwargs(func=sdq_pl.plot, kwargs=self._PARAMS)

    @property
    def _HAS_TRAIN(self):
        return self._tracker._HAS_TRAIN
    
    @property
    def _HAS_PRETRAIN(self):
        return self._tracker._HAS_PRETRAIN

    @property
    def _PLOT_KEYS(self):
        
        keys = []
        if self._HAS_PRETRAIN:
            keys+= list(self._PLOT_INPUTS["pretrain"].keys())
            
        if self._HAS_TRAIN:
            keys+= self._PLOT_INPUTS["train"].keys()
            
        return keys

    @property
    def _NPLOTS(self):
        nplots = 0        
        if self._HAS_TRAIN:
            nplots += len(self._PLOT_INPUTS["train"].keys())
        if self._HAS_PRETRAIN:
            nplots += len(self._PLOT_INPUTS["pretrain"].keys())
        print(f"nplots: {nplots}")
        return nplots

    def __layout__(self):
        
        self.fig, self.axes = sdq_pl.plot(
            nplots=self._NPLOTS,
            **self._PLOT_KWARGS,
        )
        for en, ax in enumerate(self.axes):
            ax.set_title(
                self._PLOT_TITLES[self._PLOT_KEYS[en]], fontsize=self._title_fontsize
            )
            ax.set_ylabel(
                self._PLOT_LABELS[self._PLOT_KEYS[en]], fontsize=self._label_fontsize
            )
            ax.set_xlabel(self._xlabel, fontsize=self._label_fontsize)
            ax.tick_params(axis="both", which="both", labelsize=self._tick_param_size)

    def __plot__(self):

        self.__layout__()

        self._COMBINED_PLOT_INPUTS = {}

        if self._HAS_PRETRAIN:
            self._COMBINED_PLOT_INPUTS = self._PLOT_INPUTS["pretrain"]
        if self._HAS_TRAIN:
            self._COMBINED_PLOT_INPUTS.update(self._PLOT_INPUTS["train"])

        for en, ax in enumerate(self.axes):
            val = self._COMBINED_PLOT_INPUTS[self._PLOT_KEYS[en]]
            for col_i, col in enumerate(val.columns):
                self.axes[en].plot(val[col], color=self._colors[col_i], label=col, lw=2)
                if self._grid:
                    self.axes[en].grid(**self._grid_kwargs)
            self.axes[en].legend(**self._legend_kwargs)


class VisualizeTrackedLoss(lightning.Callback):
    def __init__(
        self,
        version,
        viz_frequency = 1,
        model_name="scDiffEq_model",
        working_dir=os.getcwd(),
        train_version=0,
        pretrain_version=0,
        fname = "scDiffEq_fit_loss_tracking.png",
        *args,
        **kwargs,
    ):
        self.model_tracker = ModelTracker(
            **utils.extract_func_kwargs(func=ModelTracker, kwargs=locals())
        )
        self._INFO = utils.InfoMessage()
        self.fname = fname
        self.viz_frequency = viz_frequency
        
    @property
    def _VIZ_DISABLE(self):
        return not self.vis_frequency is None
        
    @property
    def save_path(self):
        return os.path.join(self.model_tracker._PATH, self.fname)
        
    def _save_plot(self):
        if not os.path.exists(self.save_path):
            self._INFO(f"Loss visualization saved to: {self.save_path}")
        plt.savefig(self.save_path)
        plt.close()
        
    def on_train_epoch_end(self, trianer, pl_module, *args, **kwargs):
        
        epoch = pl_module.current_epoch
        
        if not self._VIZ_DISABLED and (epoch % self.viz_frequency == 0):
            
            loss_track_viz = LossTrackingVisualization(self.model_tracker)
            loss_track_viz.__plot__()
            self._save_plot()




# -- import packages: ----------------------------------------------------------
import torch
import anndata
import autodevice
import adata_query
import numpy as np
import pandas as pd
import time
import os
import ABCParse


# -- import local dependencies: ------------------------------------------------
from ..core import utils

from ._final_state_per_simulation import FinalStatePerSimulation
from ._cell_potential import normalize_cell_potential
from ._norm import L2Norm
from ._knn import kNN


# -- set typing: ---------------------------------------------------------------
from typing import Union, List, Optional
NoneType = type(None)


class Simulator(ABCParse.ABCParse):
    """Base class for the simulator containing the core functions."""
    def __init__(
        self,
        adata: Union[anndata.AnnData, NoneType] = None,
        DiffEq = None,
        idx = None,
        use_key: str = "X_pca",
        UMAP = None,
        PCA = None,
        t_min: Union[float, NoneType] = None,
        t_max: Union[float, NoneType] = None,
        dt: float = 0.1,
        N: int = 2000,
        device: Union[torch.device, str] = "cuda:0",
        time_key: str = "Time point",
        ref_cell_type_key="Cell type annotation",
        gene_ids_key: str = "gene_ids",
        simulation_key_added: str = "simulation",
        final_state_key_added: str = "final_state",
        normalize_potential: bool = True,
        potential_normalization_kwargs={},
        time_key_added: str = "t",
        silent: bool = False,
        graph = None,
        ref_kNN_key: str = "X_pca",
        obs_mapping_keys: List[str] = ["leiden", "Cell type annotation"],
        fate_key: str = "Cell type annotation",
        return_adata: bool = True,
        name: Union[str, NoneType] = None,
        save_h5ad: bool = False,
        gpu: bool = True,
        wd = ".",
        *args,
        **kwargs,
    ):
        """ 
        adata: Union[anndata.AnnData, NoneType] = None,
        DiffEq = None,
        idx = None,
        use_key: str = "X_pca",
        UMAP = None,
        PCA = None,
        t_min: Union[float, NoneType] = None,
        t_max: Union[float, NoneType] = None,
        dt: float = 0.1,
        N: int = 2000,
        device: Union[torch.device, str] = "cuda:0",
        time_key: str = "Time point",
        ref_cell_type_key="Cell type annotation",
        gene_ids_key: str = "gene_ids",
        simulation_key_added: str = "simulation",
        final_state_key_added: str = "final_state",
        normalize_potential: bool = True,
        potential_normalization_kwargs={},
        time_key_added: str = "t",
        silent: bool = False,
        graph = None,
        ref_kNN_key: str = "X_pca",
        obs_mapping_keys: List[str] = ["leiden", "Cell type annotation"],
        fate_key: str = "Cell type annotation",
        return_adata: bool = True,
        name: Union[str, NoneType] = None,
        save_h5ad: bool = False,
        wd = ".",
        """

        self.__parse__(locals(), public=['DiffEq', 'idx'], ignore=["adata"])
        self._INFO = utils.InfoMessage(silent=silent)
        self._L2Norm = L2Norm()
        self._adata_input = adata.copy()

    @property
    def kNN(self):
        if isinstance(self._graph, NoneType):
            self._graph = kNN(self._adata_input, use_key=self._ref_kNN_key)
        return self._graph

    @property
    def _T_MIN(self):
        if isinstance(self._t_min, NoneType):
            self._t_min = self._adata_input.obs[self._time_key].min()
        return self._t_min

    @property
    def _T_MAX(self):
        if isinstance(self._t_max, NoneType):
            self._t_max = self._adata_input.obs[self._time_key].max()
        return self._t_max

    @property
    def _T_RANGE(self):
        return self._T_MAX - self._T_MIN

    @property
    def _N_STEPS(self):
        return int(self._T_RANGE / self._dt + 1)

    @property
    def t(self):
        return torch.linspace(self._T_MIN, self._T_MAX, self._N_STEPS)

    @property
    def Z0(self):
        return adata_query.fetch(
            adata = self._adata_input[self.idx],
            key=self._use_key,
            torch = self._gpu,
            device=self._device,
        ).expand(self._N, -1)
    
    @property
    def _LATENT_DIMS(self):
        return self.Z0.shape[-1]

    @property
    def _TIME(self) -> np.ndarray:
        """time for adata"""
        return np.repeat(np.linspace(self._T_MIN, self._T_MAX, self._N_STEPS), self._N)

    @property
    def _POTENTIAL_MODEL(self):
        return self.DiffEq.DiffEq.mu_potential

    @property
    def X(self):
        return self.Z_hat.flatten(0, 1).numpy()

    def _compose_core_adata(self):
        """
        Notes:
        ------
        Simulated matrix no longer sparse - no sense saving as sparse matrix.
        """
        self._INFO("Formatting as AnnData")
        self.adata = anndata.AnnData(self.X)
        self.adata.obs["t"] = self._TIME
        self.adata.uns["sim_idx"] = self.idx
        self.adata.uns[self._gene_ids_key] = self._adata_input.var[self._gene_ids_key].values

    def forward(self):
        n_nonzero_steps = int(self._N_STEPS - 1)
        self._INFO(f"Simulating {self._N} trajectories over {n_nonzero_steps} steps.")
        self.Z_hat = self.DiffEq(Z0=self.Z0, t=self.t).detach().cpu()
        self.Z_input = self.Z_hat.flatten(0, 1).to(self._device)
        self._compose_core_adata()

    def compute_drift(self):
        self._INFO("Computing per-cell drift")
        self.X_drift = self.DiffEq.DiffEq.drift(self.Z_input).detach().cpu()
        self.adata.obsm["X_drift"] = self.X_drift.numpy()
        self.adata.obs["drift"] = self._L2Norm(self.X_drift)

    def compute_diffusion(self):
        self._INFO("Computing per-cell diffusion")
        self.X_diffusion = (
            self.DiffEq.DiffEq.diffusion(self.Z_input).detach().cpu().squeeze(-1)
        )
        self.adata.obsm["X_diffusion"] = self.X_diffusion.numpy()
        self.adata.obs["diffusion"] = self._L2Norm(self.X_diffusion)

    def compute_potential(self):

        self._INFO("Computing per-cell potential")
        self.cell_potential = (
            self.DiffEq.DiffEq.mu(self.Z_input).detach().cpu().numpy().flatten()
        )
        self.adata.obs["_psi"] = self.cell_potential

        if self._normalize_potential:
            
            potential_normalization_kws={
                "q": 0.05,
                "kNN_use_key": "X",
                "knn_smoothing_iters": 5,
                "use_tqdm": False,
            }
            potential_normalization_kws.update(self._potential_normalization_kwargs)
            
            self._INFO("Normalizing per-cell potential")
            normalize_cell_potential(
                self.adata,
                use_key="_psi",
                key_added="psi",
                **potential_normalization_kws,
            )

    def map_obs(self):
        
        mapped = self.kNN.aggregate(self.adata.X, obs_keys=self._obs_mapping_keys, max_only=True)
        joined_keys = ",` `".join(self._obs_mapping_keys)
        self._INFO(f"Mapping observed label: [`{joined_keys}`] to simulated cells")
        mapped.index = mapped.index.astype(str)
        self.adata.obs = pd.concat([self.adata.obs.copy(), mapped], axis=1)
            
    def map_final_state_obs(self):
        
        df = self.adata.obs.copy()
        tf_idx = df.loc[df[self._time_key_added] == self._T_MAX].index        
        
        joined_keys = ",` `".join(self._obs_mapping_keys)
        self._INFO(f"Mapping only the final state observed label: [`{joined_keys}`]")
        
        X_final = self.adata[tf_idx].X
        mapped = self.kNN.aggregate(X_final, obs_keys=self._obs_mapping_keys, max_only=True)
        joined_keys = "_".join(self._obs_mapping_keys)
        self.adata.uns[f'mapped_final.{joined_keys}'] = mapped

    def annotate_final_state(self):

        for obs_key in self._obs_mapping_keys:
            final_state = FinalStatePerSimulation(
                adata=self.adata,
                obs_key=obs_key,
                N_sim=self._N,
                time_key=self._time_key_added,
                sim_key=self._simulation_key_added,
            )
            self._INFO(f"Annotating final state per simulation: `{obs_key}`")
            final_state.annotate_final_state(key_added=f"final_state.{obs_key}")
            

    def count_fates(self, fate_key = None):
        """
        Parameters:
        -----------
        fate_key
        """
        
        if isinstance(fate_key, NoneType):
            fate_key = self._fate_key
        
        df = self.adata.obs.copy()
        fate_counts = df.loc[df[self._time_key_added] == self._T_MAX][fate_key].value_counts()
        self.adata.uns[f"fate_counts_{fate_key}"] = fate_counts.to_dict()
        return fate_counts

    @property
    def fate_counts(self):
        """ """
        return self.count_fates(self._fate_key)
    
    def run_umap(self):
        if not isinstance(self._UMAP, NoneType):
            self._INFO("Projecting simulated cells into UMAP space")
            self._X_umap = self._UMAP.transform(self.X)
            self.adata.obsm["X_umap"] = self._X_umap

    def run_inverse_pca(self):
        if not isinstance(self._PCA, NoneType):
            self._INFO("Inverting PCA to feature space")
            self._X_gene = self._PCA.inverse_transform(self.X)
            self.adata.obsm["X_gene"] = self._X_gene

    @property
    def X_umap(self):
        if not hasattr(self, "_X_umap"):
            self.run_umap()
        return self._X_umap

    @property
    def X_gene(self):
        if not hasattr(self, "_X_gene"):
            self.run_inverse_pca()
        return self._X_gene

    @property
    def _H5AD_PATH(self):
        
        if not os.path.exists(self._wd):
            os.mkdir(self._wd)
            
        if not isinstance(self._name, NoneType):
            ID = ".".join([self._name, self.idx])
        else:
            ID = self.idx
        fname = f"adata.scDiffEq_simulated.{ID}.h5ad"
        return os.path.join(self._wd, fname)
    
    @property
    def _COMPUTE_TIME(self):
        if not hasattr(self, "_compute_time"):
            self._T_FINAL = time.time()
            self._compute_time = self._T_FINAL - self._T_INIT
        self._INFO("Compute time: {:.2f}".format(self._compute_time))

    def __call__(
        self,
        adata: Union[anndata.AnnData, NoneType] = None,
        DiffEq = None,
        idx: Union[str, int] = None,
        obs_mapping_keys: List[str] = ["leiden"],
        fate_key: str = "Cell type annotation",
        use_key: str = "X_pca",
        return_adata: bool = True,
        normalize_potential: bool = True,
        potential_normalization_kwargs={},
        name: Union[str, NoneType] = None,
        save_h5ad: bool = False,
        gene_ids_key: str = "gene_ids",
    ) -> anndata.AnnData:
        
        """Runs everything, if given."""

        # -- inputs: -----------------------------------------------------------
        self.__update__(
            locals(),
            private=[
                "use_key",
                "return_adata",
                "normalize_potential",
                "obs_mapping_keys",
                "potential_normalization_kwargs",
                "name",
                "gene_ids_key",
                "fate_key",
                "save_h5ad",
            ],
        )

        self._T_INIT = time.time()

        self._adata_input = adata.copy()

        # -- run ---------------------------------------------------------------
        self.forward()
        self.compute_drift()
        self.compute_diffusion()
        self.compute_potential()
        self.map_obs()
        self.map_final_state_obs()
        self.annotate_final_state()
        self.run_inverse_pca()
        self.run_umap()
        self.count_fates()

        # -- outputs: ----------------------------------------------------------

        if self._save_h5ad:
            self.adata.write_h5ad(self._H5AD_PATH)

        if self._return_adata:
            return self.adata

    def __repr__(self):
        return "scDiffEq Simulator"


def simulate(
    adata: anndata.AnnData,
    model,
    idx,
    obs_mapping_keys: List[str] = ["leiden", "Cell type annotation"],
    fate_key: str = "Cell type annotation",
    use_key: str = "X_pca",
    return_adata: bool = True,
    normalize_potential: bool = True,
    potential_normalization_kwargs={},
    name: Optional[str] = None,
    save_h5ad: bool = False,
    gene_ids_key: str = "gene_ids",
    t_min: Optional[float] = None,
    t_max: Optional[float] = None,
    dt: float = 0.1,
    N: int = 2000,
    device: Union[torch.device, str] = autodevice.AutoDevice(),
    time_key: str = "Time point",
    ref_cell_type_key="Cell type annotation",
    time_key_added: str = "t",
    simulation_key_added: str = "simulation",
    final_state_key_added: str = "final_state",
    silent: bool = False,
    graph=None,
    ref_kNN_key: str = "X_pca",
    PCA=None,
    UMAP=None,
    wd=".",
    return_simulator=False,
    gpu: bool = True,
    *args,
    **kwargs
) -> anndata.AnnData:

    """
    Parameters:
    -----------

    Returns:
    --------
    """

    KWARGS = utils.function_kwargs(Simulator, locals())
    KWARGS.update(kwargs)
    simulator = Simulator(**KWARGS)
    
    DiffEq = model.DiffEq
    KWARGS = utils.function_kwargs(simulator, locals())

    adata_sim = simulator(**KWARGS)

    if return_simulator:
        return adata_sim, simulator

    return adata_sim